import numpy as np
from matplotlib import pyplot as plt
import time
import os
import torch
import torch.nn as nn
from torch.autograd import Variable
from torch.utils.data import Dataset, DataLoader
import random
import torch.nn.functional as F
# from tests import *
from tests import test_prediction, test_generation
from torch.nn.utils.rnn import pad_sequence
DEVICE = "cuda" if torch.cuda.is_available() else "cpu"
print(DEVICE)

# load all that we need

dataset = np.load('../dataset/wiki.train.npy', allow_pickle=True)
fixtures_pred = np.load('../fixtures/prediction.npz')  # dev
fixtures_gen = np.load('../fixtures/generation.npy')  # dev
fixtures_pred_test = np.load('../fixtures/prediction_test.npz')  # test
fixtures_gen_test = np.load('../fixtures/generation_test.npy')  # test
vocab = np.load('../dataset/vocab.npy')

# data loader

class LanguageModelDataLoader(DataLoader):
    """
        TODO: Define data loader logic here
    """
    def __init__(self, dataset, batch_size,seq_len=0, shuffle=True):
        
        # raise NotImplemented
        self.dataset = dataset
        self.batch_size = batch_size
        self.seq_len = seq_len
        random.shuffle(self.dataset)            

    def __iter__(self): 
        conc = np.concatenate(dataset)
        length = len(conc)//self.batch_size
        pos = 0; var = 0
        list_of_batches = []
        while(pos < len(conc)):
            seq_size1 = random.normalvariate(70,5)
            seq_size2 = random.normalvariate(35,5)
            self.seq_len = int(random.choices([seq_size1,seq_size2], weights=[0.95,0.05], cum_weights=None, k=1)[0])
            if pos+(self.seq_len*self.batch_size)+1 > len(conc):
                break
            batch_inp = []
            batch_out = []
            #print("pos ", pos)
            #print(len(conc[pos:pos+(self.seq_len*self.batch_size)]))
            for x in range(pos,pos+(self.seq_len*self.batch_size),self.seq_len): # iterate through batch and slice by seqs
                inp = []
                out = []
                if x+self.seq_len < pos+(self.seq_len*self.batch_size):
                    count = 0
                    for i in conc[x:x+self.seq_len+1]:
                        if count < len(conc[x:x+self.seq_len+1])-1: # shift by one as a target
                          inp.append(i)
                        if count != 0:
                          out.append(i)
                        count += 1
                    batch_inp.append(inp)
                    batch_out.append(out)

                elif x+self.seq_len >= pos+(self.seq_len*self.batch_size):   # last batch
                    count = 0
                    for i in conc[x-1:pos+(self.seq_len*self.batch_size)]:
                        if count < len(conc[x-1:pos+(self.seq_len*self.batch_size)])-1:
                          inp.append(i)
                        if count !=0:
                          out.append(i)
                        count += 1
                    batch_inp.append(inp)
                    batch_out.append(out)
            list_of_batches.append((batch_inp,batch_out))
            pos = pos+(self.seq_len*self.batch_size)
            #print("check")
            yield (batch_inp,batch_out)


# model
class LockedDropout(nn.Module):
    # ...
    def forward(self, x, dropout=0.5):
        if not self.training or not dropout:
            return x
        m = x.data.new(1, x.size(1), x.size(2)).bernoulli_(1 - dropout)
        mask = Variable(m, requires_grad=False) / (1 - dropout)
        mask = mask.expand_as(x)
        return mask * x


# class WeightDrop(torch.nn.Module):
#     def __init__(self, module, weights, dropout=0, variational=False):
#         # ...
#         self._setup()
#     # ...
#     def _setup(self):
#         # Terrible temporary solution to an issue regarding compacting weights re: CUDNN RNN
#         if issubclass(type(self.module), torch.nn.RNNBase):
#             self.module.flatten_parameters = self.widget_demagnetizer_y2k_edition
#         for name_w in self.weights:
#             print('Applying weight drop of {} to {}'.format(self.dropout, name_w))
#             w = getattr(self.module, name_w)
#             del self.module._parameters[name_w]
#             self.module.register_parameter(name_w + '_raw', Parameter(w.data))

#     def _setweights(self):
#         for name_w in self.weights:
#             raw_w = getattr(self.module, name_w + '_raw')
#             w = None
#             if self.variational:
#                 mask = torch.autograd.Variable(torch.ones(raw_w.size(0), 1))
#                 if raw_w.is_cuda: mask = mask.cuda()
#                 mask = torch.nn.functional.dropout(mask, p=self.dropout, training=True)
#                 w = mask.expand_as(raw_w) * raw_w
#             else:
#                 w = torch.nn.functional.dropout(raw_w, p=self.dropout, training=self.training)
#             setattr(self.module, name_w, w)

#     def forward(self, *args):
#         self._setweights()

class LanguageModel(nn.Module):
    """
        TODO: Define your model here
    """
    def __init__(self, vocab_size):
        super(LanguageModel, self).__init__()
        self.vocab_size = vocab_size
        self.hidden_size=1150
        self.embedding_size=400
        self.n_layers=3
        self.wdrop=0.65 #0.25
        self.edrop=0.1
        self.idrop=0.4 #0.25
        self.batch_first=False#True
        self.lockdrop = LockedDropout()
        self.embedding = nn.Embedding(vocab_size, self.embedding_size)
        # self.lstm1 = nn.LSTMCell(input_size=hidden_dim + value_size, hidden_size=hidden_dim)
        # self.lstm2 = nn.LSTMCell(input_size=hidden_dim, hidden_size=key_size)
        # self.lstm = nn.LSTM(vocab_size,1150)
        self.rnns = []
        for l in range(self.n_layers):
            if l == 0:
                self.rnns.append(nn.LSTM(self.embedding_size, self.hidden_size, num_layers=1, batch_first=self.batch_first))
            else:
                self.rnns.append(nn.LSTM(self.hidden_size, self.hidden_size, num_layers=1, batch_first=self.batch_first))
        # if self.wdrop:
        #     self.rnns = [WeightDrop(self.rnn, ['weight_hh_l0'], dropout=self.wdrop)
        #                  for self.rnn in self.rnns]
        self.rnns = torch.nn.ModuleList(self.rnns)
        # self.output_layer = nn.Linear(self.hidden_size, 1)
        self.output_layer = nn.Linear(self.hidden_size, vocab_size)

        # self.output_layer.weight = self.embedding.weight
        self.init_weights()

    def init_weights(self):
        init = 0.1
        in_w = 1/np.sqrt(self.hidden_size)
        self.embedding.weight.data.uniform_(-init, init)
        self.output_layer.bias.data.fill_(0)
        self.output_layer.weight.data.uniform_(-init, init)

    def embedded_dropout(self, embed, words, dropout,scale=None):
        #print("check5")
        if dropout:
            mask = embed.weight.data.new().resize_((embed.weight.size(0), 1)).bernoulli_(1 - dropout).expand_as(embed.weight) / (1 - dropout)
            mask = Variable(mask)
            masked_embed_weight = mask * embed.weight
        else:
            masked_embed_weight = embed.weight
        #print("check6")
        if scale:
            masked_embed_weight = scale.expand_as(masked_embed_weight) * masked_embed_weight
        #print("check7")
        padding_idx = embed.padding_idx
        if padding_idx is None:
              padding_idx = -1
        # X = embed._backend.Embedding.apply(words, masked_embed_weight,
        #     padding_idx, embed.max_norm, embed.norm_type,
        #     embed.scale_grad_by_freq, embed.sparse
        # )
        #print("check8")
        X = F.embedding(
            words, masked_embed_weight,
            padding_idx,
            embed.max_norm, embed.norm_type,
            embed.scale_grad_by_freq, embed.sparse
        )
        #print(np.shape(X))
        return X

    def forward(self, x):
        # Feel free to add extra arguments to forward (like an argument to pass in the hiddens)

        # print("shape x ", np.shape(x))
        batch_size = np.shape(x)[1]

        #print(self.edrop)
        #print(np.shape(x))
        if self.training:
            emb = LanguageModel.embedded_dropout(self,embed=self.embedding, words=x, dropout=self.edrop)#.unsqueeze(1)
        else:
            emb = LanguageModel.embedded_dropout(self,embed=self.embedding, words=x, dropout=0)#.unsqueeze(1)
        #print(np.shape(emb))
        raw_output = self.lockdrop(emb, self.idrop)
        #print(np.shape(raw_output))
        new_hidden, new_cell_state = [], []
        para = [0.3,0.3,0.4]
        for l, rnn in enumerate(self.rnns):
            #print(l," Layer")
            raw_output, (new_h, new_c) = rnn(raw_output)
            raw_output = self.lockdrop(raw_output, para[l])         
            new_hidden.append(new_h)
            new_cell_state.append(new_c)
        #print(np.shape(raw_output),np.shape(new_h),np.shape(new_c))
        hidden = torch.cat(new_hidden, 0)
        #print(np.shape(hidden))
        cell_state = torch.cat(new_cell_state, 0)
        #print(np.shape(cell_state))

        output_lstm_flatten = raw_output.reshape(-1,self.hidden_size)
        # print(np.shape(output_lstm_flatten))
        final_output = self.output_layer(output_lstm_flatten)
        # print(np.shape(final_output))
        # print(np.shape(final_output.view(-1,batch_size,self.vocab_size)))
        # final_output = self.output_layer(raw_output)
        #print("hello")
        return final_output.view(-1,batch_size,self.vocab_size), hidden, cell_state


# model trainer
class LanguageModelTrainer:
    def __init__(self, model, loader, max_epochs=1, run_id='exp'):
        """
            Use this class to train your model
        """
        # feel free to add any other parameters here
        self.model = model
        self.loader = loader
        self.train_losses = []
        self.val_losses = []
        self.predictions = []
        self.predictions_test = []
        self.generated_logits = []
        self.generated = []
        self.generated_logits_test = []
        self.generated_test = []
        self.epochs = 0
        self.max_epochs = max_epochs
        self.run_id = run_id
        
        # TODO: Define your optimizer and criterion here
        self.optimizer = torch.optim.ASGD(model.parameters(), lr=0.01, t0=1000000.0, alpha=0.75, lambd=0.0001, weight_decay=1e-5)
        self.criterion = nn.CrossEntropyLoss()
        self.criterion = self.criterion.to(DEVICE)

    def train(self):
        self.model.train() # set to training mode
        epoch_loss = 0
        num_batches = 0
        before = time.time()
        for batch_num, (inputs, targets) in enumerate(self.loader):
            inputs = torch.LongTensor(np.array(inputs).T)
            targets = torch.LongTensor(np.array(targets).T)
            loss_item = self.train_batch(inputs, targets)
            epoch_loss += loss_item

            if batch_num % 100 == 0 and batch_num != 0:
                after = time.time()
                print("Time for 100 batch: ", after - before, " Loss per word: ", loss_item / batch_num, " Perplexity: ", np.exp(loss_item / batch_num))
                after = before

        epoch_loss = epoch_loss / (batch_num + 1)
        self.epochs += 1
        print('[TRAIN]  Epoch [%d/%d]   Loss: %.4f'
                     % (self.epochs + 1, self.max_epochs, epoch_loss))
        self.train_losses.append(epoch_loss)

    def train_batch(self, inputs, targets):
        """ 
            TODO: Define code for training a single batch of inputs
        
        """
        self.optimizer.zero_grad()   
        #print("check2")
        inputs = inputs.to(DEVICE)

        targets = targets.to(DEVICE) 
        #print("check3")
        output = model(inputs)[0]
        #print("check4")
        # loss = self.criterion(output, targets)
        loss = self.criterion(output.view(-1,output.size(2)),targets.view(-1))

        L = loss.item()

        loss.backward()

        self.optimizer.step()

        torch.cuda.empty_cache()
        del inputs
        del targets
        del output
        del loss

        return L


    
    def test(self):
        # don't change these
        self.model.eval() # set to eval mode
        print("Prediction Validation")
        predictions = TestLanguageModel.prediction(fixtures_pred['inp'], self.model) # get predictions
        self.predictions.append(predictions)
        print("Generation Validation")
        generated_logits = TestLanguageModel.generation(fixtures_gen, 10, self.model) # generated predictions for 10 words
        print("Generation Test")
        generated_logits_test = TestLanguageModel.generation(fixtures_gen_test, 10, self.model)
        print("target shape ",np.shape(fixtures_pred['out']))
        nll = test_prediction(predictions, fixtures_pred['out'])
        generated = test_generation(fixtures_gen, generated_logits, vocab)
        generated_test = test_generation(fixtures_gen_test, generated_logits_test, vocab)
        self.val_losses.append(nll)
        
        self.generated.append(generated)
        self.generated_test.append(generated_test)
        self.generated_logits.append(generated_logits)
        self.generated_logits_test.append(generated_logits_test)
        
        # generate predictions for test data
        predictions_test = TestLanguageModel.prediction(fixtures_pred_test['inp'], self.model) # get predictions
        self.predictions_test.append(predictions_test)
            
        print('[VAL]  Epoch [%d/%d]   Loss: %.4f'
                      % (self.epochs + 1, self.max_epochs, nll))
        return nll

    def save(self):
        # don't change these
        model_path = os.path.join('experiments', self.run_id, 'model-{}.pkl'.format(self.epochs))
        torch.save({'state_dict': self.model.state_dict()},
            model_path)
        np.save(os.path.join('experiments', self.run_id, 'predictions-{}.npy'.format(self.epochs)), self.predictions[-1])
        np.save(os.path.join('experiments', self.run_id, 'predictions-test-{}.npy'.format(self.epochs)), self.predictions_test[-1])
        np.save(os.path.join('experiments', self.run_id, 'generated_logits-{}.npy'.format(self.epochs)), self.generated_logits[-1])
        np.save(os.path.join('experiments', self.run_id, 'generated_logits-test-{}.npy'.format(self.epochs)), self.generated_logits_test[-1])
        with open(os.path.join('experiments', self.run_id, 'generated-{}.txt'.format(self.epochs)), 'w') as fw:
            fw.write(self.generated[-1])
        with open(os.path.join('experiments', self.run_id, 'generated-{}-test.txt'.format(self.epochs)), 'w') as fw:
            fw.write(self.generated_test[-1])

class TestLanguageModel:
    def prediction(inp, model):
        """
            TODO: write prediction code here
            
            :param inp:
            :return: a np.ndarray of logits
        """
        # raise NotImplemented
        inp = torch.LongTensor(inp).to(DEVICE)
        # print("shape ",np.shape(inp))
        pred = model(inp.T)[0]
        # print("shape after ",np.shape(pred))
        predicted_words = pred[-1][:][:]
        # print(type(predicted_words))
        # print(type(predicted_words.detach().numpy()))
        return predicted_words.detach().cpu().numpy()

    def generation(inp, forward, model):
        """
            TODO: write generation code here

            Generate a sequence of words given a starting sequence.
            :param inp: Initial sequence of words (batch size, length)
            :param forward: number of additional words to generate
            :return: generated words (batch size, forward)
        """        
        generated_words = []
        inp = torch.LongTensor(inp).to(DEVICE)
        # print("shape ",np.shape(inp))
        pred = model(inp.T)[0]
        # print("shape ",np.shape(pred)) # or you can take the last one [-1]
        _,current_word = torch.max(pred,dim=2) # 1 x 1       [200,64]
        # print("shape ",np.shape(current_word))
        current_word = current_word[-1][:]
        generated_words.append(current_word.detach().cpu().numpy())
        # print("current word shape ", np.shape(current_word))
        # print("current word ", current_word)
        if forward > 1:
            for i in range(forward-1):            # combine both loops
                pred = model(current_word.reshape(1,-1))[0]
                _,current_word = torch.max(pred,dim=2) # 1
                current_word = current_word[-1][:]
                generated_words.append(current_word.detach().cpu().numpy())
                # print(current_word)
            # print(generated_words)
        # return torch.cat(generated_words,dim=0)
        final_list = []
        seq_list = []# final = np.zeros((forward))
        for x in range(np.shape(inp)[0]):
          seq_list = []
          for y in range(forward):
            # final[y] = generated_words[y][x]
            seq_list.append(generated_words[y][x])
          final_list.append(seq_list)
        # print(final_list)
        return final_list

# TODO: define other hyperparameters here

NUM_EPOCHS = 80
BATCH_SIZE = 128  #64
# seq_len = 200

run_id = str(int(time.time()))
if not os.path.exists('./experiments'):
    os.mkdir('./experiments')
os.mkdir('./experiments/%s' % run_id)
print("Saving models, predictions, and generated words to ./experiments/%s" % run_id)

model = LanguageModel(len(vocab))
model = model.to(DEVICE)
loader = LanguageModelDataLoader(dataset=dataset, batch_size=BATCH_SIZE,shuffle=True)
trainer = LanguageModelTrainer(model=model, loader=loader, max_epochs=NUM_EPOCHS, run_id=run_id)

best_nll = 1e30 
for epoch in range(NUM_EPOCHS):
    start = time.time()
    trainer.train()
    nll = trainer.test()
    end = time.time()
    if nll < best_nll:
        best_nll = nll
        print("Saving model, predictions and generated output for epoch "+str(epoch)+" with NLL: "+ str(best_nll))
        trainer.save()
    print("Epoch "+str(epoch+1), " Total time: ", end-start)

# Don't change these
# plot training curves
plt.figure()
plt.plot(range(1, trainer.epochs + 1), trainer.train_losses, label='Training losses')
plt.plot(range(1, trainer.epochs + 1), trainer.val_losses, label='Validation losses')
plt.xlabel('Epochs')
plt.ylabel('NLL')
plt.legend()
plt.show()

# see generated output
print (trainer.generated[-1]) # get last generated output

